@@ -156,27 +156,17 @@ int main(int argc, char *argv[])
 	}
 
 	const char* filename = argv[1];
-	WPXInputStream* input = new libwpg::WPGFileStream(filename);
-	if (input->isOLEStream())
-	{
-		WPXInputStream* olestream = input->getDocumentOLEStream();
-		if (olestream)
-		{
-			delete input;
-			input = olestream;
-		}
-	}
+	libwpg::WPGFileStream input(filename);
 
-	if (!libwpg::WPGraphics::isSupported(input))
+	if (!libwpg::WPGraphics::isSupported(&input))
 	{
 		fprintf(stderr, "ERROR: Unsupported file format (unsupported version) or file is encrypted!\n");
 		return 1;
 	}
 
 	RawPainter painter;
-	libwpg::WPGraphics::parse(input, &painter);
+	libwpg::WPGraphics::parse(&input, &painter);
 
-	delete input;
 	return 0;
 }
 
