@@ -157,20 +157,33 @@ bool WPG1Parser::parse()
 
 	static const struct RecordHandler handlers[] =
 	{
-		{ 0x01, "Fill Attributes",   &WPG1Parser::handleFillAttributes },	 
-		{ 0x02, "Line Attributes",   &WPG1Parser::handleLineAttributes },	 
-		{ 0x03, "Marker Atttibutes", 0 },	 
-		{ 0x04, "Polymarker",        0 },	 
-		{ 0x05, "Line",              &WPG1Parser::handleLine },	 
-		{ 0x06, "Polyline",          &WPG1Parser::handlePolyline },	 
-		{ 0x07, "Rectangle",         &WPG1Parser::handleRectangle },	 
-		{ 0x08, "Polygon",           &WPG1Parser::handlePolygon },	 
-		{ 0x09, "Ellipse",           &WPG1Parser::handleEllipse },	 
-		{ 0x0b, "Bitmap (Type 1)",   &WPG1Parser::handleBitmapTypeOne },	 
-		{ 0x0e, "Colormap",          &WPG1Parser::handleColormap },
-		{ 0x0f, "Start WPG",         &WPG1Parser::handleStartWPG },
-		{ 0x10, "End WPG",           &WPG1Parser::handleEndWPG },
-		{ 0x14, "Bitmap (Type 2)",   &WPG1Parser::handleBitmapTypeTwo },	 
+		{ 0x01, "Fill Attributes",           &WPG1Parser::handleFillAttributes },	 
+		{ 0x02, "Line Attributes",           &WPG1Parser::handleLineAttributes },	 
+		{ 0x03, "Marker Atttibutes",         0 },	 
+		{ 0x04, "Polymarker",                0 },	 
+		{ 0x05, "Line",                      &WPG1Parser::handleLine },	 
+		{ 0x06, "Polyline",                  &WPG1Parser::handlePolyline },	 
+		{ 0x07, "Rectangle",                 &WPG1Parser::handleRectangle },	 
+		{ 0x08, "Polygon",                   &WPG1Parser::handlePolygon },	 
+		{ 0x09, "Ellipse",                   &WPG1Parser::handleEllipse },	 
+		{ 0x0a, "Reserved",                  0 },	 
+		{ 0x0b, "Bitmap (Type 1)",           &WPG1Parser::handleBitmapTypeOne },	 
+		{ 0x0c, "Graphics Text (Type 1)",    0 },	 
+		{ 0x0d, "Grephics Text Attributes",  0 },	 
+		{ 0x0e, "Colormap",                  &WPG1Parser::handleColormap },
+		{ 0x0f, "Start WPG",                 &WPG1Parser::handleStartWPG },
+		{ 0x10, "End WPG",                   &WPG1Parser::handleEndWPG },
+		{ 0x11, "Postscript Data (Type 1)",  &WPG1Parser::handlePostscriptTypeOne },	 
+		{ 0x12, "Output Attributes",         0 },	 
+		{ 0x13, "Curved Polyline",           &WPG1Parser::handleCurvedPolyline },	 
+		{ 0x14, "Bitmap (Type 2)",           &WPG1Parser::handleBitmapTypeTwo },	 
+		{ 0x15, "Start Figure",              0 },	 
+		{ 0x16, "Start Chart",               0 },	 
+		{ 0x17, "Planperfect Data",          0 },	 
+		{ 0x18, "Graphics Text (Type 2)",    0 },	 
+		{ 0x19, "Start WPG (Type 2)",        0 },	 
+		{ 0x1a, "Graphics Text (Type 3)",    0 },	 
+		{ 0x1b, "Postscript Data (Type 2)",  &WPG1Parser::handlePostscriptTypeTwo },	 
 		{ 0x00, 0, 0 } // end marker
 	};	 
 
@@ -283,6 +296,8 @@ void WPG1Parser::handleColormap()
 		return;
 	unsigned startIndex = readU16();
 	unsigned numEntries = readU16();
+	if (startIndex > 255 || numEntries > 256 || startIndex + numEntries > 256)
+		return;
 
 	WPG_DEBUG_MSG(("Colormap\n"));
 	for(unsigned int i = 0; i < numEntries; i++)
@@ -325,7 +340,10 @@ void WPG1Parser::handleLineAttributes()
 
 	m_pen.solid = style != 0;
 	m_pen.foreColor = m_colorPalette[color];
-	m_pen.width = (double)width / 1200.0;
+	if (!width && m_pen.solid)
+		m_pen.width = 0.001;
+	else
+		m_pen.width = (double)width / 1200.0;
 
 	WPG_DEBUG_MSG(("Line Attributes\n"));
 	WPG_DEBUG_MSG(("         Line style: %d\n", style));
@@ -387,7 +405,8 @@ void WPG1Parser::handleRectangle()
 
 	libwpg::WPGRect rect;
 	rect.x1 = (double)x/1200.0;
-	rect.y1 = (double)(m_height-y)/1200.0;
+	// in WPG, we have the coordinate of lower left point, in SVG-ish coordinates we have to get upper left
+	rect.y1 = (double)(m_height - h - y)/1200.0;
 	rect.x2 = rect.x1 + (double)w/1200.0;
 	rect.y2 = rect.y1 + (double)h/1200.0;
 
@@ -430,6 +449,10 @@ void WPG1Parser::handleEllipse()
 	int cy = readS16();
 	int rx = readS16();
 	int ry = readS16();
+	int rotation = readS16();
+	int beginAngle = readS16();
+	int endAngle = readS16();
+	unsigned flags = readU16();
 
 	libwpg::WPGPoint center;
 	center.x = (double)cx/1200.0;
@@ -448,31 +471,57 @@ void WPG1Parser::handleEllipse()
 	WPG_DEBUG_MSG(("        Radius y: %d\n", ry));
 }
 
-// allocate a buffer (of specified bitmap size), then start decoding RLE data
-// from the input stream to fill the buffer
-// NOTE: delete the buffer by yourself!
-unsigned char* WPG1Parser::decodeRLE(int width, int height, int depth)
+void WPG1Parser::handleCurvedPolyline()
 {
-	if (depth <= 0 || width <= 0 || height <= 0)
-		return 0;
+	if (!m_graphicsStarted)
+		return;
+	readU32();
+	unsigned int count = readU16();
+	if (!count)
+		return;
+	libwpg::WPGPath path;
+	path.closed = false;
+	long xInitial = readS16();
+	long yInitial = readS16();
+	path.moveTo(libwpg::WPGPoint((double)xInitial/1200.0, (double)(m_height-yInitial)/1200.0));
+	for (unsigned i = 1; i < (count-1)/3; i++)
+	{
+		long xControl1 = readS16();
+		long yControl1 = readS16();
+		long xControl2 = readS16();
+		long yControl2 = readS16();
+		long xCoordinate = readS16();
+		long yCoordinate = readS16();
+		path.curveTo(libwpg::WPGPoint((double)xControl1/1200.0, (double)(m_height-yControl1)/1200.0),
+			libwpg::WPGPoint((double)xControl2/1200.0, (double)(m_height-yControl2)/1200.0),
+			libwpg::WPGPoint((double)xCoordinate/1200.0, (double)(m_height-yCoordinate)/1200.0));
+	}
+	
+	m_painter->setBrush(path.closed ? m_brush : libwpg::WPGBrush());
+	m_painter->setPen(m_pen);
+	m_painter->drawPath(path);
+	
+	WPG_DEBUG_MSG(("Curved Polyline\n"));
+}
+
+void WPG1Parser::decodeRLE(std::vector<unsigned char>& buffer, unsigned width, unsigned height, unsigned depth)
+{
+	buffer.clear();
+
+	// This are the known depth values for WPG1, no point to try to decode others since they are likely to indicate corruption
+	if (depth != 8 && depth != 4 && depth != 2 && depth != 1)
+		return;
 
 	// round to the next byte
 	unsigned scanline_width = (width * depth + 7)/8;
-	unsigned buffer_size = scanline_width * height;
+	unsigned tmpBufferSize = scanline_width * height;
 	WPG_DEBUG_MSG(("Scanline width: %d\n", scanline_width));
-	WPG_DEBUG_MSG(("Output size: %d\n", buffer_size));
-
-	// buffer for the result
-	unsigned char* buffer = new unsigned char[buffer_size];
-	
-	if (!buffer)
-		return 0;
-
-	// pointer to write decoded data
-	unsigned char* ptr = buffer;
+	WPG_DEBUG_MSG(("Output size: %d\n", scanline_width * height));
 
 	WPG_DEBUG_MSG(("Decoding RLE data\n"));
-	for(; m_input->tell() < m_recordEnd; )
+
+	buffer.reserve(tmpBufferSize);
+ 	while (m_input->tell() < m_recordEnd && !m_input->atEOS() && buffer.size() < tmpBufferSize)
 	{
 		unsigned char opcode = readU8();
 
@@ -483,8 +532,8 @@ unsigned char* WPG1Parser::decodeRLE(int width, int height, int depth)
 			unsigned char pixel = (count > 0) ? readU8() : 0xff;
 			if(count == 0)
 				count = (int)readU8();
-			for( ; count && ptr < buffer + buffer_size; --count)
-				*ptr++ = pixel;
+			for( ; count ; --count)
+				buffer.push_back( pixel );
 		}
 		else
 		{
@@ -492,34 +541,44 @@ unsigned char* WPG1Parser::decodeRLE(int width, int height, int depth)
 			if(count > 0)
 			{
 				// literal byte copy
-				for( ; count && ptr < buffer + buffer_size; --count)
-					*ptr++ = readU8();
+				for( ; count ; --count)
+					buffer.push_back( readU8() );
 			}
 			else
 			{
 				// copy entire scan line
 				count = (int)readU8();
-				unsigned char* raster_source = ptr - scanline_width;
-				if (!raster_source || raster_source < buffer)
-					return 0;
+				if (buffer.size() < scanline_width )
+				{
+					WPG_DEBUG_MSG(("Cannot copy the scanline, not enough data %li\n", (long)buffer.size()));
+					break;
+				}
+				unsigned raster_source = buffer.size() - scanline_width;
 				for( ; count; --count)
-					for(unsigned r = 0; r < scanline_width && ptr < buffer + buffer_size; r++)
-						*ptr++ = raster_source[r];
+					for(unsigned r = 0; r < scanline_width ; r++)
+					{
+						unsigned char pixel = buffer[raster_source + r];
+						buffer.push_back( pixel );
+					}
 			}
 		}
 	}
 	WPG_DEBUG_MSG(("Finish decoding RLE data\n"));
-
-	return buffer;
+	WPG_DEBUG_MSG(("Buffer length: %li\n", (long)buffer.size()));
+	
+	while (buffer.size() < tmpBufferSize)
+		buffer.push_back(0);
 }
 
-void WPG1Parser::fillPixels(libwpg::WPGBitmap& bitmap, unsigned char* buffer, 
-int width, int height, int depth)
+void WPG1Parser::fillPixels(libwpg::WPGBitmap& bitmap, const unsigned char* buffer, unsigned width, unsigned height, unsigned depth)
 {
 	// sanity
 	if(!buffer)
 		return;
 
+	if (depth != 8 && depth != 4 && depth != 2 && depth != 1)
+		return;
+
 	// round to the next byte
 	unsigned scanline_width = (width * depth + 7)/8;
 
@@ -528,24 +587,51 @@ int width, int height, int depth)
 	{
 		libwpg::WPGColor black(0, 0, 0);
 		libwpg::WPGColor white(255, 255, 255);
-		for(int y = 0; y < height; y++)
+		for(unsigned y = 0; y < height; y++)
 		{
-			unsigned char* buf = buffer + y * scanline_width;
-			for(int x = 0; x < width; x++)
+			const unsigned char* buf = buffer + y * scanline_width;
+			for(unsigned x = 0; x < width; x++)
 				if(buf[x/8] & (0x80 >> (x & 7)))
 					bitmap.setPixel(x, y, white);
 				else
 					bitmap.setPixel(x, y, black);
 		}
 	}
-
-	// 8-bit image: indexed color
-	if(depth == 8)
+	// 2-bit image: 4-color bitmap (indexed)
+	else if(depth == 2)
 	{
-		for(int y = 0; y < height; y++)
+		unsigned i = 0;
+		for (unsigned y = 0; y < height; y++)
+		for (unsigned x = 0; x < width; x++, i++)
 		{
-			unsigned char* buf = buffer + y * scanline_width;
-			for(int x = 0; x < width; x++)
+			if ((x==0) && (i % 4 != 0))
+				i = (i/4 + 1) * 4;
+			unsigned index = ((buffer[i/4] & (0x03 << 2*(3 - (i % 4)))) >> 2*(3 - (i % 4)));
+			const libwpg::WPGColor& color = m_colorPalette[index];
+			bitmap.setPixel(x, y, color);
+		}
+	}
+	// 4 -bit image: 16-colour bitmap (indexed)
+	else if(depth == 4)
+	{
+		unsigned i = 0;
+		for (unsigned y = 0; y < height; y++)
+		for (unsigned x = 0; x < width; x++, i++)
+		{
+			if ((x==0) && (i % 2 != 0))
+				i = (i/2 + 1) * 2;
+			unsigned index = ((buffer[i/2] & (0x0f << 4*(1 - (i % 2)))) >> 4*(1 - (i % 2)));
+			const libwpg::WPGColor& color = m_colorPalette[index];
+			bitmap.setPixel(x, y, color);
+		}
+	}
+	// 8-bit image: 256-colour image (indexed)
+	else if(depth == 8)
+	{
+		for(unsigned y = 0; y < height; y++)
+		{
+			const unsigned char* buf = buffer + y * scanline_width;
+			for(unsigned x = 0; x < width; x++)
 			{
 				const libwpg::WPGColor& color = m_colorPalette[buf[x]];
 				bitmap.setPixel(x, y, color);
@@ -557,8 +643,8 @@ int width, int height, int depth)
 	// debugging only
 	if(buffer && 0)
 	{
-		for(int x = 0; x < width; x++)
-		for(int y = 0; y < height; y++)
+		for(unsigned x = 0; x < width; x++)
+		for(unsigned y = 0; y < height; y++)
 		{
 			libwpg::WPGColor color = bitmap.pixel(x,y);
 			WPG_DEBUG_MSG((" pixel at %d, %d: %3d %3d %3d\n", x, y, color.red, color.green, color.blue));
@@ -583,6 +669,10 @@ void WPG1Parser::handleBitmapTypeOne()
 	WPG_DEBUG_MSG(("Horizontal resolution: %d\n", hres));
 	WPG_DEBUG_MSG(("  Vertical resolution: %d\n", vres));
 
+	// if this happens, likely corruption, bail out.
+	if (depth != 1 && depth != 2 && depth != 4 && depth != 8)
+		return;
+		
 	// Sanity checks
 	if(hres <= 0)
 		hres = 1200;
@@ -603,12 +693,11 @@ void WPG1Parser::handleBitmapTypeOne()
 	bitmap.rect.x2 = (double)width/(double)hres;
 	bitmap.rect.y2 = (double)height/(double)vres;
 
-	unsigned char* buffer = decodeRLE(width, height, depth);
-	if (buffer)
+	std::vector<unsigned char> buffer;
+	decodeRLE(buffer, width, height, depth);
+	if (buffer.size() && buffer.size() == (size_t)((width*depth + 7)/8)*height)
 	{
-		fillPixels(bitmap, buffer, width, height, depth);
-		delete [] buffer;
-
+		fillPixels(bitmap, &buffer[0], width, height, depth);
 		m_painter->drawBitmap(bitmap);
 	}
 }
@@ -617,11 +706,7 @@ void WPG1Parser::handleBitmapTypeTwo()
 {
 	if (!m_graphicsStarted)
 		return;
-#ifdef DEBUG
 	int rotation = readS16();
-#else
-	readS16();
-#endif
 	int x1 = readS16();
 	int y1 = readS16();
 	int x2 = readS16();
@@ -642,6 +727,12 @@ void WPG1Parser::handleBitmapTypeTwo()
 	WPG_DEBUG_MSG(("Horizontal resolution: %d\n", hres));
 	WPG_DEBUG_MSG(("  Vertical resolution: %d\n", vres));
 
+	// if this happens, likely corruption, bail out.
+	if (rotation < 0 || rotation > 359)
+		return;
+	if (depth != 1 && depth != 2 && depth != 4 && depth != 8)
+		return;
+		
 	// Sanity checks
 	if(hres <= 0)
 		hres = 1200;
@@ -669,16 +760,85 @@ void WPG1Parser::handleBitmapTypeTwo()
 	bitmap.rect.x2 = (double)xs2/1200.0;
 	bitmap.rect.y2 = (double)(ys2)/1200.0;
 
-	unsigned char* buffer = decodeRLE(width, height, depth);
-	if (buffer)
+	std::vector<unsigned char> buffer;
+	decodeRLE(buffer, width, height, depth);
+	if (buffer.size() && buffer.size() == (size_t)((width*depth + 7)/8)*height)
 	{
-		fillPixels(bitmap, buffer, width, height, depth);
-		delete [] buffer;
-
+		fillPixels(bitmap, &buffer[0], width, height, depth);
 		m_painter->drawBitmap(bitmap);
 	}
 }
 
+void WPG1Parser::handlePostscriptTypeOne()
+{
+	if (!m_graphicsStarted)
+		return;
+	long x1 = readS16();
+	long y1 = readS16();
+	long x2 = readS16();
+	long y2 = readS16();
+	libwpg::WPGBinaryData data;
+	data.rect.x1 = (double)x1/72.0;
+	data.rect.y1 = (double)m_height/1200.0 - (double)y1/72.0;
+	data.rect.x2 = (double)x2/72.0;
+	data.rect.y2 = (double)m_height/1200.0 - (double)y2/72.0;
+	
+	data.clear();
+	while (m_input->tell() <= m_recordEnd)
+		data.append((char)readU8());
+	data.mimeType = "application/x-postscript";
+	if (data.size())
+		m_painter->drawImageObject(data);
+}
+
+void WPG1Parser::handlePostscriptTypeTwo()
+{
+	if (!m_graphicsStarted)
+		return;
+#ifdef DEBUG
+	unsigned lengthOfData = readU32();
+	int rotation = readS16();
+#else
+	readU32();
+	readS16();
+#endif
+	int x1 = readS16();
+	int y1 = readS16();
+	int x2 = readS16();
+	int y2 = readS16();
+
+	WPG_DEBUG_MSG(("Postscript (Type 2)\n"));
+	WPG_DEBUG_MSG(("       Length of Data: %d\n", lengthOfData));
+	WPG_DEBUG_MSG(("       Rotation Angle: %d\n", rotation));
+	WPG_DEBUG_MSG(("   Bottom left corner: %d,%d\n", x1, y1));
+	WPG_DEBUG_MSG(("     Top right corner: %d,%d\n", x2, y2));
+
+	y1 = m_height - y1;
+	y2 = m_height - y2;
+
+	long xs1 = (x1 <= x2) ? x1 : x2;
+	long xs2 = (x1 <= x2) ? x2 : x1;
+	long ys1 = (y1 <= y2) ? y1 : y2;
+	long ys2 = (y1 <= y2) ? y2 : y1;
+	WPG_DEBUG_MSG(("%li %li   %li %li\n", xs1, ys1, xs2, ys2));
+
+	libwpg::WPGBinaryData data;
+	data.rect.x1 = (double)xs1/1200.0;
+	data.rect.y1 = (double)(ys1)/1200.0;
+	data.rect.x2 = (double)xs2/1200.0;
+	data.rect.y2 = (double)(ys2)/1200.0;
+	
+	data.mimeType = "image/x-eps";
+
+	m_input->seek(48, WPX_SEEK_CUR);
+	
+	data.clear();
+	while (m_input->tell() <= m_recordEnd)
+		data.append((char)readU8());
+	if (data.size())
+		m_painter->drawImageObject(data);
+}
+
 void WPG1Parser::resetPalette()
 {
 	m_colorPalette.clear();
