@@ -31,12 +31,15 @@
 
 #include <math.h>
 #include <vector>
+#include <sstream>
 
 // MSVC++ 6.0 does not have the macro defined, so we define it
 #ifndef M_PI
 #define M_PI 3.14159265358979323846
 #endif
 
+#define DUMP_BINARY_DATA 0
+
 static const unsigned char defaultWPG2PaletteRed[] = {
 	0x00, 0xFF, 0x7F, 0xBF, 0x00, 0x00, 0x00, 0x7F,
 	0x7F, 0x7F, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF,
@@ -165,9 +168,9 @@ public:
 	long kxsin;
 	long kysin;
 	long txinteger;
-	short txfraction;
+	unsigned short txfraction;
 	long tyinteger;
-	short tyfraction;
+	unsigned short tyfraction;
 	long px;
 	long py;
 
@@ -227,7 +230,8 @@ WPG2Parser::WPG2Parser(WPXInputStream *input, libwpg::WPGPaintInterface* painter
 	m_bitmap(),
 	m_binaryData(),
 	m_hFlipped(false),
-	m_vFlipped(false)
+	m_vFlipped(false),
+	m_binaryId(0)
 {
 }
 
@@ -262,7 +266,7 @@ bool WPG2Parser::parse()
 		{ 0x11, "Chart Data",           0 },     // ignored
 		{ 0x12, "Object Image",         &WPG2Parser::handleObjectImage },
 		{ 0x15, "Polyline",             &WPG2Parser::handlePolyline },
-		{ 0x16, "Polyspline",           0 },
+		{ 0x16, "Polyspline",           &WPG2Parser::handlePolyspline },
 		{ 0x17, "Polycurve",            &WPG2Parser::handlePolycurve },
 		{ 0x18, "Rectangle",            &WPG2Parser::handleRectangle },
 		{ 0x19, "Arc",                  &WPG2Parser::handleArc },
@@ -1192,6 +1196,15 @@ void WPG2Parser::handlePolyline()
 		WPG_DEBUG_MSG(("        Point #%d : %g,%g\n", j+1, points[j].x, points[j].y));
 }
 
+void WPG2Parser::handlePolyspline()
+{
+	if (!m_graphicsStarted)
+		return;
+	// We have never seen a polyspline in a WPG file and we are not even able to generate
+	// one, so, we draw a polyline instead and wait for bug reports :)
+	handlePolyline();
+}
+
 void WPG2Parser::handlePolycurve()
 {
 	if (!m_graphicsStarted)
@@ -1397,14 +1410,14 @@ void WPG2Parser::handleBitmapData()
 {
 	if (!m_graphicsStarted)
 		return;
-	unsigned long width = readU16();
-	unsigned long height = readU16();
+	unsigned width = readU16();
+	unsigned height = readU16();
 	unsigned color_format = readU8();
 	unsigned compression_format = readU8();
 
 	WPG_DEBUG_MSG(("     dimension : %g, %g  %g, %g\n", m_bitmap.x1, m_bitmap.y1, m_bitmap.x2, m_bitmap.y2));
-	WPG_DEBUG_MSG(("         width : %li pixels\n", width));
-	WPG_DEBUG_MSG(("        height : %li pixels\n", height));
+	WPG_DEBUG_MSG(("         width : %i pixels\n", width));
+	WPG_DEBUG_MSG(("        height : %i pixels\n", height));
 	WPG_DEBUG_MSG(("  color format : %d\n", color_format));
 	WPG_DEBUG_MSG(("   compression : %d (%s)\n", compression_format,
 		(compression_format==0) ? "uncompressed":
@@ -1412,8 +1425,8 @@ void WPG2Parser::handleBitmapData()
 
 	if (color_format > 12) // not documented and we are likely not knowing what to do with this
 		return;
-	unsigned long tmpBufferSize;
-	unsigned long raster_len = width;
+	unsigned tmpBufferSize;
+	unsigned raster_len = width;
 	if (color_format == 1)
 	{
 		tmpBufferSize = (width/8+1)*height;
@@ -1443,31 +1456,25 @@ void WPG2Parser::handleBitmapData()
 		tmpBufferSize = width*height;
 	
 	
-	// prepare the bitmap actual data
-	unsigned char* buffer = 0;
-	unsigned char* ptr = 0;
+	std::vector<unsigned char> buffer;
+	buffer.reserve(tmpBufferSize);
 
 	// plain data, uncompression
 	if(compression_format==0)
-	{
-		buffer = new unsigned char[tmpBufferSize];
 		for(unsigned ii=0; ii < tmpBufferSize && m_input->tell() <= m_recordEnd; ii++)
-			buffer[ii] = readU8();
-	}
+			buffer.push_back( readU8() );
 	// run-length encoding
 	else if(compression_format==1)
 	{
 		unsigned char data[256];
 		bool xor_raster = false;
-		unsigned char* next_scanline = buffer;
+		unsigned next_scanline = 0;
 		unsigned data_size = 1;
-		buffer = new unsigned char[tmpBufferSize];
-		ptr = buffer;
 
 		WPG_DEBUG_MSG(("Decoding RLE data\n"));
 
 		// FIXME check for ptr, it should not go out of bound!!
-		while (m_input->tell() <= m_recordEnd)
+		while (m_input->tell() <= m_recordEnd && !m_input->atEOS() && buffer.size() < tmpBufferSize)
 		{
 			unsigned char opcode = readU8();
 			// specify data size
@@ -1480,14 +1487,10 @@ void WPG2Parser::handleBitmapData()
 					if (tmpBufferSize < data_size*width*height)
 					{
 						tmpBufferSize = data_size*width*height;
-						if (buffer)
-							delete [] buffer;
-						buffer = new unsigned char[tmpBufferSize];
+						buffer.reserve(tmpBufferSize);
 					}
 					raster_len = data_size*width;
 				}
-				ptr = buffer;
-				next_scanline = buffer;
 			}
 
 			// a run of black (index #0)
@@ -1495,8 +1498,8 @@ void WPG2Parser::handleBitmapData()
 			{
 				unsigned count = 1 + readU8();
 				for( ; count ; --count )
-					for(unsigned j = 0; j < data_size; j++)
-						*ptr++ =  0;
+					for(unsigned j = 0; j < data_size && !m_input->atEOS(); j++)
+						buffer.push_back( 0 );
 				
 			}
 
@@ -1506,8 +1509,8 @@ void WPG2Parser::handleBitmapData()
 				unsigned count = 1 + readU8();
 
 				for( ; count ; --count )
-					for(unsigned j = 0; j < data_size; j++)
-						*ptr++ =  255;
+					for(unsigned j = 0; j < data_size && !m_input->atEOS(); j++)
+						buffer.push_back( 255 );
 			}
 
 			// extend previous run
@@ -1515,18 +1518,23 @@ void WPG2Parser::handleBitmapData()
 			{
 				unsigned count = 1 + readU8();
 				for( ; count; --count)
-					for(unsigned j = 0; j < data_size; j++)
-						*ptr++ = data[j];
+					for(unsigned j = 0; j < data_size && !m_input->atEOS(); j++)
+						buffer.push_back( data[j] );
 			}
 
 			// repeat raster
 			else if(opcode == 0xfe)
 			{
 				unsigned count = 1 + readU8();
-				unsigned char* raster_source = ptr - raster_len;
+				if ( buffer.size() < raster_len )
+					break;
+				unsigned raster_source = buffer.size() - raster_len;
 				for( ; count; --count)
 					for(unsigned long r = 0; r < raster_len; r++)
-						*ptr++ = *raster_source++;
+					{
+						unsigned char pixel = buffer[raster_source + r];
+						buffer.push_back( pixel );
+					}
 			}
 
 			// XOR
@@ -1535,7 +1543,7 @@ void WPG2Parser::handleBitmapData()
 				// Xor-ing will happen when about to enter next raster
 				// see after the last if down below
 				xor_raster = true;
-				next_scanline = ptr + raster_len;
+				next_scanline = buffer.size() + raster_len;
 			}
 
 			// NOTE: the following two IFs here must be the last ones
@@ -1544,11 +1552,11 @@ void WPG2Parser::handleBitmapData()
 			else if(opcode >= 0x80)
 			{
 				unsigned count = 1 + (opcode & 0x7f);
-				for(unsigned i = 0; i < data_size; i++)
+				for(unsigned i = 0; i < data_size && !m_input->atEOS(); i++)
 					data[i] = readU8();
 				for( ; count; --count)
 					for(unsigned j = 0; j < data_size; j++)
-						*ptr++ = data[j];
+						buffer.push_back( data[j] );
 			}
 
 			// simple copy
@@ -1556,8 +1564,8 @@ void WPG2Parser::handleBitmapData()
 			{
 				unsigned count = opcode + 1;
 				for( ; count; --count)
-					for(unsigned j = 0; j < data_size; j++)
-						*ptr++ = readU8();
+					for(unsigned j = 0; j < data_size && !m_input->atEOS(); j++)
+						buffer.push_back( readU8() );
 			}
 
 			// unreachable: only sentinel
@@ -1568,16 +1576,18 @@ void WPG2Parser::handleBitmapData()
 			}
 
 			// new raster/scanline? if necessary do the XOR now
-			if(xor_raster && (ptr >= next_scanline))
+			if(xor_raster && buffer.size() >= next_scanline)
 			{
 				// reset, because XOR in one raster at a time
 				xor_raster = false;
 
 				// current raster must be XORed with previous raster
-				unsigned char* current = next_scanline - raster_len;
-				unsigned char* previous = current - raster_len;
-				for( unsigned long r = 0; r < raster_len; r++)
-					current[r] ^= previous[r];
+				unsigned current = next_scanline - raster_len;
+				unsigned previous = current - raster_len;
+				if (current >= buffer.size() || previous >= buffer.size())
+					return;
+				for( unsigned r = 0; r < raster_len; r++)
+					buffer[current++] ^= buffer[previous++];
 
 			}
 
@@ -1587,7 +1597,10 @@ void WPG2Parser::handleBitmapData()
 	} 
 	
 	// no buffer? format is unknown 
-	if(!buffer) return;
+	if(!buffer.size()) return;
+	
+	while (buffer.size() < tmpBufferSize)
+		buffer.push_back(0);
 
 	// prepare the bitmap structure for the listener
 	libwpg::WPGBitmap bitmap(width, height, m_vFlipped, m_hFlipped);
@@ -1737,7 +1750,7 @@ void WPG2Parser::handleBitmapData()
 	
 #ifdef DEBUG
 	// debugging only
-	if(buffer && 0)
+	if(buffer.size() && 0)
 	{
 		for(unsigned long x = 0; x < width; x++)
 		for(unsigned long y = 0; y < height; y++)
@@ -1747,9 +1760,6 @@ void WPG2Parser::handleBitmapData()
 		}
 	}
 #endif
-
-	// discard
-	delete [] buffer;
 }
 
 void WPG2Parser::handleObjectCapsule()
@@ -1772,10 +1782,10 @@ void WPG2Parser::handleObjectCapsule()
 	long ys1 = (y1 <= y2) ? y1 : y2;
 	long ys2 = (y1 <= y2) ? y2 : y1;
 
-	m_binaryData.x1 = TO_DOUBLE(xs1) / 1200;
-	m_binaryData.y1 = TO_DOUBLE(ys1) / 1200;
-	m_binaryData.x2 = TO_DOUBLE(xs2) / 1200;
-	m_binaryData.y2 = TO_DOUBLE(ys2) / 1200;
+	m_binaryData.x1 = TO_DOUBLE(xs1) / m_xres;
+	m_binaryData.y1 = TO_DOUBLE(ys1) / m_yres;
+	m_binaryData.x2 = TO_DOUBLE(xs2) / m_xres;
+	m_binaryData.y2 = TO_DOUBLE(ys2) / m_yres;
 	
 	unsigned long numDescriptions = readU16();
 	
@@ -1785,8 +1795,8 @@ void WPG2Parser::handleObjectCapsule()
 	static const char* mimeTypesMap[] =
 	{
 		"", // 0
-		"image/x-wpg",
-		"image/x-wpg",
+		"", // "image/x-wpg",
+		"", // "image/x-wpg",
 		"",
 		"",
 		"", // 5
@@ -1830,7 +1840,7 @@ void WPG2Parser::handleObjectCapsule()
 		unsigned char description = readU8();
 		if (description < 0x27)
 			m_binaryData.mimeTypes.push_back(libwpg::WPGString(mimeTypesMap[description]));
-		m_input->seek(2, WPX_SEEK_CUR);
+		m_input->seek(7, WPX_SEEK_CUR);
 	}
 
 	m_binaryData.objectIndex = 0;
@@ -1864,13 +1874,18 @@ void WPG2Parser::handleObjectImage()
 	WPG_DEBUG_MSG(("     Image Object Size : %li\n", (unsigned long)binaryData.size()));
 
 	// temporary for debug - dump the binary data (need to have write access in the current directory
-#if 0
-	FILE *f = fopen("binarydump", "w");
-	const char *tmpBinaryBuffer = binaryData.getDataBuffer();
-	const size_t tmpBufferSize = binaryData.size();
-	for (size_t k = 0; k < tmpBufferSize; k++)
-		fprintf(f, "%c",tmpBinaryBuffer[k]);
-	fclose(f);
+#if DUMP_BINARY_DATA
+	std::ostringstream filename;
+	filename << "binarydump" << m_binaryId++ << ".bin";
+	FILE *f = fopen(filename.str().c_str(), "wb");
+	if (f)
+	{
+		const char *tmpBinaryBuffer = binaryData.getDataBuffer();
+		const size_t tmpBufferSize = binaryData.size();
+		for (size_t k = 0; k < tmpBufferSize; k++)
+			fprintf(f, "%c",tmpBinaryBuffer[k]);
+		fclose(f);
+	}
 #endif
 
 
