@@ -175,6 +175,7 @@ private:
 	void handleBrushPattern();
 
 	void handlePolyline();
+	void handlePolyspline();
 	void handlePolycurve();
 	void handleRectangle();
 	void handleArc();
@@ -221,6 +222,7 @@ private:
 
 	class ObjectCharacterization;
 	void parseCharacterization(ObjectCharacterization*);
+	unsigned m_binaryId;
 };
 
 #endif // __WPG2PARSER_H__
