@@ -38,7 +38,7 @@
 #include <string>
 #include <vector>
 
-// #include "libwpg_utils.h"
+#include <string.h>
 
 namespace libwpg
 {
