@@ -40,10 +40,14 @@ WPGXParser::WPGXParser(const WPGXParser& parser):
 
 unsigned char WPGXParser::readU8()
 {
+	if (!m_input || m_input->atEOS())
+		return (unsigned char)0;
 	size_t numBytesRead;
 	unsigned char const * p = m_input->read(sizeof(unsigned char), numBytesRead);
 	
-	return *(unsigned char const *)(p);
+	if (p && numBytesRead == 1)
+		return *(unsigned char const *)(p);
+	return (unsigned char)0;
 }
 
 unsigned short WPGXParser::readU16()
@@ -53,37 +57,28 @@ unsigned short WPGXParser::readU16()
 	return (unsigned short)(p0|(p1<<8));
 }
 
-unsigned long WPGXParser::readU32()
+unsigned int WPGXParser::readU32()
 {
-	unsigned long p0 = (unsigned short)readU8();
-	unsigned long p1 = (unsigned short)readU8();
-	unsigned long p2 = (unsigned short)readU8();
-	unsigned long p3 = (unsigned short)readU8();
+	unsigned int p0 = (unsigned int)readU8();
+	unsigned int p1 = (unsigned int)readU8();
+	unsigned int p2 = (unsigned int)readU8();
+	unsigned int p3 = (unsigned int)readU8();
 	return (unsigned long)(p0|(p1<<8)|(p2<<16)|(p3<<24));
 }
 
-char WPGXParser::readS8()
+signed char WPGXParser::readS8()
 {
-	size_t numBytesRead;
-	unsigned char const * p = m_input->read(sizeof(unsigned char), numBytesRead);
-	
-	return *(signed char const *)(p);
+	return (signed char)readU8();
 }
 
 short WPGXParser::readS16()
 {
-	short p0 = readU8();
-	short p1 = readS8();
-	return (short)(p0|(p1<<8));
+	return (short)readU16();
 }
 
-long WPGXParser::readS32()
+int WPGXParser::readS32()
 {
-	long p0 = readU8();
-	long p1 = readU8();
-	long p2 = readU8();
-	long p3 = readS8();
-	return (long)(p0|(p1<<8)|(p2<<16)|(p3<<24));
+	return (int)readU32();
 }
 
 unsigned int WPGXParser::readVariableLengthInteger()
